/* -*- Mode: C; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 8 -*- */

/*
 *  Copyright (C) 2003 Hiroyuki Ikezoe
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2, or (at your option)
 *  any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <stdlib.h>
#include <string.h>

#include "gobject-utils.h"
#include "intl.h"
#include "kz-http.h"
#include "kazehakase.h"
#include "kz-profile.h"

#define BUFFER_SIZE 1024

enum {
	START_SIGNAL,
	FETCH_COMPLETED_SIGNAL,
	CONNECT_ERROR_SIGNAL,
	LAST_SIGNAL
};

enum {
	PROP_0,
	PROP_METHOD,
	PROP_URI,
	PROP_HOSTNAME,
	PROP_PATH,
};

static const struct {
	const gchar *name;
	KzHTTPMethodType type;
} methods[] = {
	{"HEAD", KZ_HTTP_METHOD_HEAD},
	{"GET",  KZ_HTTP_METHOD_GET},
};

static guint n_methods = G_N_ELEMENTS(methods);

static void kz_http_class_init      (KzHTTPClass *klass);
static void kz_http_init            (KzHTTP *http);
static void kz_http_set_property    (GObject *object,
				     guint prop_id,
				     const GValue *value,
				     GParamSpec *pspec);
static void kz_http_get_property    (GObject *object,
				     guint prop_id,
				     GValue *value,
				     GParamSpec *pspec);

static void      kz_http_connect         (KzHTTP *http);
static void      kz_http_error           (KzHTTP *http);
static GIOStatus kz_http_in_header       (KzHTTP *http,
					  GIOChannel *iochannel);
static GIOStatus kz_http_in_body         (KzHTTP *http,
					  GIOChannel *iochannel);
static GIOStatus kz_http_in_chunked_body (KzHTTP *http,
					  GIOChannel *iochannel);

static void     cb_http_connect  (GTcpSocket *socket, 
				  GTcpSocketConnectAsyncStatus status,
				  gpointer data);
static gboolean cb_http_in       (GIOChannel *iochannel,
				  GIOCondition condition,
				  gpointer data);

static GObjectClass *parent_class = NULL;
static gint kz_http_signals[LAST_SIGNAL] = {0};

KZ_OBJECT_GET_TYPE(kz_http, "KzHTTP", KzHTTP,
		   kz_http_class_init, kz_http_init,
		   G_TYPE_OBJECT)

static void
kz_http_class_init (KzHTTPClass *klass)
{
	GObjectClass *object_class;

	parent_class = g_type_class_peek_parent (klass);
	object_class = (GObjectClass *) klass;

	object_class->dispose = kz_http_dispose;
	object_class->set_property = kz_http_set_property;
	object_class->get_property = kz_http_get_property;
	
	klass->fetch_completed     = NULL;
	klass->connect_error       = NULL;
	
	g_object_class_install_property(
		object_class,
		 PROP_URI,
		 g_param_spec_string(
			 "uri",
			 _("URI"),
			 _("The URI of Fetch file"),
			 NULL,
			 G_PARAM_READWRITE));
	g_object_class_install_property(
		object_class,
		 PROP_HOSTNAME,
		 g_param_spec_string(
			 "hostname",
			 _("Hostname"),
			 _("The Hostname of the URI"),
			 NULL,
			 G_PARAM_READWRITE));
	g_object_class_install_property(
		object_class,
		 PROP_PATH,
		 g_param_spec_string(
			 "path",
			 _("Path"),
			 _("The Path of the URI"),
			 NULL,
			 G_PARAM_READWRITE));

	kz_http_signals[FETCH_COMPLETED_SIGNAL]
		= g_signal_new ("fetch_completed",
				G_TYPE_FROM_CLASS (klass),
				G_SIGNAL_RUN_LAST,
				G_STRUCT_OFFSET (KzHTTPClass, fetch_completed),
				NULL, NULL,
				g_cclosure_marshal_VOID__VOID,
				G_TYPE_NONE, 0);
	kz_http_signals[CONNECT_ERROR_SIGNAL]
		= g_signal_new ("connect_error",
				G_TYPE_FROM_CLASS (klass),
				G_SIGNAL_RUN_LAST,
				G_STRUCT_OFFSET (KzHTTPClass, connect_error),
				NULL, NULL,
				g_cclosure_marshal_VOID__VOID,
				G_TYPE_NONE, 0);
}


static void
kz_http_init (KzHTTP *http)
{
	http->iochannel       = NULL;
	http->socket          = NULL;
	
	http->uri             = NULL;
	http->hostname        = NULL;
	http->path            = NULL;
	
	http->header          = TRUE;
	http->chunked         = FALSE;
	http->redirection     = FALSE;
	
	http->buf             = NULL;

	http->content_length  = 0;
	http->bytes_loaded    = 0;
	http->chunk_size      = 0;
}


void
kz_http_dispose (GObject *object)
{
	KzHTTP *http;

	g_return_if_fail(KZ_IS_HTTP(object));

	http = KZ_HTTP(object);
	
	if (http->iochannel && http->iochannel->ref_count > 0)
		g_io_channel_unref(http->iochannel);
	if (http->uri)
		g_free(http->uri);
	if (http->hostname)
		g_free(http->hostname);
	if (http->path)
		g_free(http->path);
	if (http->buf)
		g_string_free(http->buf, TRUE);
	
	http->iochannel = NULL;
	http->uri       = NULL;
	http->hostname  = NULL;
	http->path      = NULL;
	http->buf       = NULL;

	if (G_OBJECT_CLASS (parent_class)->dispose)
		G_OBJECT_CLASS (parent_class)->dispose(object);
}


static void
kz_http_set_property (GObject *object,
		      guint prop_id,
		      const GValue *value,
		      GParamSpec *pspec)
{
	KzHTTP *http = KZ_HTTP(object);

	switch (prop_id)
	{
	case PROP_URI:
		g_free(http->uri);
		http->uri = g_value_dup_string(value);
		break;
	case PROP_HOSTNAME:
		g_free(http->hostname);
		http->hostname = g_value_dup_string(value);
		break;
	case PROP_PATH:
		g_free(http->path);
		http->path = g_value_dup_string(value);
		break;
	default:
		G_OBJECT_WARN_INVALID_PROPERTY_ID(object, prop_id, pspec);
		break;
	}
}


static void
kz_http_get_property (GObject *object,
		      guint prop_id,
		      GValue *value,
		      GParamSpec *pspec)
{
	KzHTTP *http = KZ_HTTP(object);

	switch (prop_id)
	{
	case PROP_URI:
		g_value_set_string(value, http->uri);
		break;
	case PROP_HOSTNAME:
		g_value_set_string(value, http->hostname);
		break;
	case PROP_PATH:
		g_value_set_string(value, http->path);
		break;
	default:
		G_OBJECT_WARN_INVALID_PROPERTY_ID(object, prop_id, pspec);
		break;
	}
}


KzHTTP *
kz_http_new (const gchar *uri)
{
	KzHTTP *http;
	GURI *guri;
	gchar *path = NULL, *hostname = NULL;

	guri = gnet_uri_new(uri);

	if (guri)
	{
		hostname = guri->hostname;
		if (guri->query)
			path = g_strdup_printf("%s?%s",
					       guri->path, guri->query);
		else
			path = g_strdup(guri->path);
	}

	
	http = KZ_HTTP(g_object_new(KZ_TYPE_HTTP,
				    "uri",      uri,
				    "hostname", hostname,
				    "path",     path,
				    NULL));
	
	if (guri)
		gnet_uri_delete(guri);
	g_free(path);

	return http;
}


static void
kz_http_connect (KzHTTP *http)
{
	GURI *uri = NULL;
	gchar proxy[1024];
	gboolean exist;
	guint port = 80;
	gchar *host;

	/* proxy check */
	exist = kz_profile_get_value(kz_global_profile,
				     "Global", "proxy",
				     proxy, sizeof(proxy),
				     KZ_PROFILE_VALUE_TYPE_STRING);
	if (exist)
	{
		uri = gnet_uri_new(proxy);
		http->use_proxy = TRUE;
		port = uri->port;
		host = uri->hostname;
	}
	else 
	{
		host = http->hostname;
	}

	gnet_tcp_socket_connect_async(host, port,
				      cb_http_connect, http);
	
	if(uri)
		gnet_uri_delete(uri);
}


static void
kz_http_error (KzHTTP *http)
{
	g_return_if_fail(KZ_IS_HTTP(http));

	g_signal_emit(G_OBJECT(http),
		      kz_http_signals[CONNECT_ERROR_SIGNAL],
		      0);
	kz_http_dispose(G_OBJECT(http));
}


static GIOStatus
kz_http_in_header(KzHTTP *http, GIOChannel *iochannel)
{
	GIOStatus iostatus;
	GString *buffer = NULL;
	
	buffer = g_string_sized_new(0);
	/* Read the data into our buffer */
	iostatus = g_io_channel_read_line_string(iochannel,
						 buffer,
						 NULL,
						 NULL);
	
	if (iostatus == G_IO_STATUS_ERROR)
		return iostatus;
	
	if (strncmp(buffer->str, "HTTP/1.1", 8) == 0)
	{
		switch (buffer->str[9])
		{
		 case '2': /* 2xx Succeccful */
			break;
		 case '3': /* 3xx Redirection   */
			http->redirection = TRUE;
			break;
		 case '1': /* 1xx Informational */
		 case '4': /* 4xx Client Error  */
		 case '5': /* 5xx Server Error  */
		 default:
			g_warning("%s", buffer->str);
			break;
		}
	}
	else if (strncmp(buffer->str, "Content-Length:", 15) == 0)
	{
		http->content_length = strtol(buffer->str + 15, NULL, 10);
		http->buf = g_string_sized_new(http->content_length);
	}
	else if (strncmp(buffer->str, "Transfer-Encoding:", 18) == 0)
	{
		if(strncasecmp(buffer->str + 19, "chunked", 7) == 0)  
			http->chunked = TRUE;
		http->buf = g_string_sized_new(0);
	}
	else if (strncmp(buffer->str, "Location:", 9) == 0)
	{
		http->location = g_strdup(buffer->str + 10);
		if (http->redirection)
		{
			g_warning("Redirection:%s", http->location);
		}
	}
	else if (strncmp(buffer->str, "Last-Modified:", 15) == 0)
	{
		strptime(buffer->str + 15,
			 "%a, %d %b %Y %H:%M:%S %z", &(http->last_modified));
	}
	else if (strncmp(buffer->str, "\r\n", 2) == 0)
	{
		http->header = FALSE;
	}

	g_string_free(buffer, TRUE);

	return iostatus;
}


static GIOStatus
kz_http_in_chunked_body(KzHTTP *http, GIOChannel *iochannel)
{
	GIOStatus iostatus = G_IO_STATUS_NORMAL;
	GError *e = NULL;
	gchar *buffer = NULL;
	gsize length;

	/* These codes are quite silly! FIX!! */
	/* Get chunk size */
	if (http->chunk_size <= 0)
	{
		iostatus = g_io_channel_read_line(iochannel,
						  &buffer,
						  &length,
						  NULL,
						  &e);
		if (iostatus == G_IO_STATUS_NORMAL)
		{
			http->chunk_size = strtol(buffer, NULL, 16);
			g_free(buffer);
			/* check EOF */
			if (http->chunk_size == 0)
				iostatus = G_IO_STATUS_EOF;
		}
	}
	if (iostatus == G_IO_STATUS_NORMAL)
	{
		/* Get a chunked body */
		buffer = g_new0(gchar, http->chunk_size);
		iostatus = g_io_channel_read_chars(iochannel, buffer,
						   http->chunk_size,
						   &length,
						   &e);
		if (iostatus == G_IO_STATUS_NORMAL)
		{
			g_string_append_len(http->buf, buffer, length);
			http->bytes_loaded += length;
			http->chunk_size   -= length;
		}
		g_free(buffer);
	}
	if (iostatus == G_IO_STATUS_NORMAL)
	{
		/* Get CRLF */
		if (http->chunk_size <= 0)
		{
			iostatus = g_io_channel_read_line(iochannel,
							  &buffer,
							  &length,
							  NULL,
							  &e);
			g_free(buffer);
		}
	}

	if (e)
		g_error_free(e);

	return iostatus;
}


static GIOStatus
kz_http_in_body(KzHTTP *http, GIOChannel *iochannel)
{
	GIOStatus iostatus;
	guint bytes_read;
	gchar buf[BUFFER_SIZE];
	
	/* Read the data into our buffer */
	iostatus = g_io_channel_read_chars(iochannel, buf, 
					   sizeof(buf),
					   &bytes_read,
					   NULL);

	if (iostatus == G_IO_STATUS_NORMAL)
	{	
		g_string_append_len(http->buf, buf, bytes_read);
		http->bytes_loaded += bytes_read;
		/* Check for EOF */
		if (bytes_read == 0)
			iostatus = G_IO_STATUS_EOF;
	}
	return iostatus;
}


static gboolean
cb_http_in(GIOChannel *iochannel, GIOCondition condition,
	   gpointer data)
{
	KzHTTP *http;
	GIOStatus iostatus;
	
	http = KZ_HTTP(data);
	
	if ((condition & G_IO_ERR) || (condition & G_IO_HUP))
	{
		g_warning("IO Condition: %d", condition);
		kz_http_error(http);
		return FALSE;
	}
			
	if (http->header) /* Header Section */
		iostatus = kz_http_in_header(http, iochannel);
	else if (http->chunked) /* Chunked Body */
		iostatus = kz_http_in_chunked_body(http, iochannel);
	else 		/* Entity-Body Section */
		iostatus = kz_http_in_body(http, iochannel);
	
	if (iostatus == G_IO_STATUS_EOF)
	{
		g_signal_emit(G_OBJECT(http),
			      kz_http_signals[FETCH_COMPLETED_SIGNAL],
			      0);
		return FALSE;
	}
	else if (iostatus == G_IO_STATUS_NORMAL)
		return TRUE;
	else 
	{
		kz_http_error(http);
		return FALSE;
	}
}


static void 
cb_http_connect(GTcpSocket *socket,
		GTcpSocketConnectAsyncStatus status, gpointer data)
{
	KzHTTP *http;
	GIOStatus iostatus;
	gchar *command;
	gchar *method;
	guint n, i;
	gchar *URL;

	http = KZ_HTTP(data);

	if ( status != GTCP_SOCKET_CONNECT_ASYNC_STATUS_OK )
	{
		kz_http_error(http);
		return;
	}
	http->socket = socket;
	
	/* Get the IOChannel */
	http->iochannel = gnet_tcp_socket_get_io_channel(socket);
	if (http->iochannel == NULL)
	{
		kz_http_error(http);
		return;
	}

	g_io_channel_set_flags(http->iochannel, G_IO_FLAG_NONBLOCK, NULL);

	/* proxy */
	if (http->use_proxy)
	{
		URL = g_strdup_printf("http://%s%s",
				      http->hostname,
				      http->path);
	}
	else
	{
		URL = g_strdup(http->path);
	}
	
	/* Set method */
	for (i = 0; i< n_methods; i++)
	{
		if (http->method == methods[i].type)
		{
			method = methods[i].name;
			break;
		}
	}
		
	/* Send GET command */
	command = g_strconcat(method, " ",
			      URL, " HTTP/1.1\r\n",
			      "Host: ", http->hostname, "\r\n",
			      "User-Agent: Kazehakase/"VERSION "\r\n",
			      "Connection: close\r\n\r\n", NULL);

	iostatus = g_io_channel_write_chars(http->iochannel, command,
					    strlen(command), &n, NULL);

	g_free(command);
	g_free(URL);

	if (iostatus != G_IO_STATUS_NORMAL)
		kz_http_error(http);

	g_io_channel_set_buffered(http->iochannel, TRUE);

	g_io_add_watch(http->iochannel,
		       G_IO_IN | G_IO_PRI | G_IO_ERR | G_IO_HUP,
		       cb_http_in, http);
}


guint
kz_http_get_lastmodified (KzHTTP *http)
{
	g_return_val_if_fail(KZ_IS_HTTP(http), 0);

	return http->last_modified;
}


gsize
kz_http_get_file_size (KzHTTP *http)
{
	g_return_val_if_fail(KZ_IS_HTTP(http), 0);

	return http->content_length;
}


gsize
kz_http_get_loaded_size (KzHTTP *http)
{
	g_return_val_if_fail(KZ_IS_HTTP(http), 0);

	return http->bytes_loaded;
}


void
kz_http_stop_loading (KzHTTP *http)
{
	g_return_if_fail(KZ_IS_HTTP(http));
	
	if (http->iochannel)
	{
		g_io_channel_unref(http->iochannel);
		http->iochannel = NULL;
	}
	
	if (http->socket)
	{
		gnet_tcp_socket_delete(http->socket);
		http->socket = NULL;
	}
}


void
kz_http_request_head (KzHTTP *http)
{
	http->method = KZ_HTTP_METHOD_HEAD;

	kz_http_connect(http);
}


void
kz_http_file_fetch(KzHTTP *http)
{
	http->method = KZ_HTTP_METHOD_GET;

	kz_http_connect(http);
}

