/* -*- Mode: C; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 8 -*- */

/*
 *  Copyright (C) 2003 Takuro Ashie
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2, or (at your option)
 *  any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include "prefs_tab.h"

#include "kazehakase.h"
#include "utils/intl.h"

#define DATA_KEY "KzPrefsTab::info"


static GtkWidget     *prefs_tab_create          (void);
static void           prefs_tab_response        (GtkWidget *widget,
						 gint       response);


static KzPrefsWinPageEntry prefs_entry =
{
	path:            N_("/Tab"),
	priority_hint:   0,
	create:          prefs_tab_create,
	response:        prefs_tab_response,
};


typedef struct _KzPrefsTab
{
	GtkWidget     *main_vbox;
	GtkWidget     *fix_width_check;
	GtkWidget     *width_spin;
	GtkWidget     *show_close_button;
	GtkWidget     *show_progressbar;
	gboolean       changed;
} KzPrefsTab;


static void
cb_fix_width_toggled(GtkToggleButton *button, KzPrefsTab *prefsui)
{
	gtk_widget_set_sensitive(prefsui->width_spin,
				 gtk_toggle_button_get_active(button));
	prefsui->changed = TRUE;
}


static void
cb_changed (GtkWidget *widget, KzPrefsTab *prefsui)
{
	prefsui->changed = TRUE;
}


static void
prefs_tab_destroy (gpointer data)
{
	KzPrefsTab *prefsui = data;

	g_free(prefsui);
}


static GtkWidget *
prefs_tab_create (void)
{
	KzPrefsTab *prefsui = g_new0(KzPrefsTab, 1);
	GtkWidget *main_vbox, *vbox, *hbox, *frame;
	GtkWidget *label, *check, *spin;
	GtkAdjustment *adj;
	gboolean fix, show;
	gint width;

	main_vbox = gtk_vbox_new(FALSE, 0);
	prefsui->main_vbox = main_vbox;
	g_object_set_data_full(G_OBJECT(main_vbox), DATA_KEY,
			       prefsui, prefs_tab_destroy);

	label = kz_prefs_ui_utils_create_title(_("Tab"));
	gtk_box_pack_start(GTK_BOX(main_vbox), label,
			   FALSE, FALSE, 0);
	gtk_widget_show(label);

	/*
	 *  tab width
	 */
	frame = gtk_frame_new(_("Width"));
	gtk_frame_set_label_align(GTK_FRAME(frame), 0.03, 0.5);
	gtk_container_set_border_width(GTK_CONTAINER(frame), 5);
	gtk_box_pack_start(GTK_BOX(main_vbox), frame, FALSE, FALSE, 2);
	gtk_widget_show(frame);

	hbox = gtk_hbox_new(FALSE, 0);
	gtk_container_set_border_width(GTK_CONTAINER(hbox), 4);
	gtk_container_add(GTK_CONTAINER(frame), hbox);
	gtk_widget_show(hbox);

	fix = TRUE;
	kz_profile_get_value(kz_global_profile,
			     "Tab", "fix_width",
			     &fix, sizeof(fix),
			     KZ_PROFILE_VALUE_TYPE_BOOL);
	width = 80;
	kz_profile_get_value(kz_global_profile,
			     "Tab", "fixed_width",
			     &width, sizeof(width),
			     KZ_PROFILE_VALUE_TYPE_INT);

	check = gtk_check_button_new_with_mnemonic(_("_Fix tab width"));
	prefsui->fix_width_check = check;
	gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(check), fix);
	gtk_box_pack_start(GTK_BOX(hbox), check, FALSE, FALSE, 2);
	g_signal_connect(G_OBJECT(check), "toggled",
			 G_CALLBACK(cb_fix_width_toggled), prefsui);
	gtk_widget_show(check);

	prefsui->width_spin = spin
		= gtk_spin_button_new_with_range(-1, 1000, 1);
	adj = gtk_spin_button_get_adjustment(GTK_SPIN_BUTTON(spin));
	gtk_adjustment_set_value(adj, width);
	gtk_box_pack_start(GTK_BOX(hbox), spin, FALSE, FALSE, 2);
	g_signal_connect(G_OBJECT(spin), "value-changed",
			 G_CALLBACK(cb_changed), prefsui);
	gtk_widget_show(spin);

	/*
	 *  Shiw/Hide
	 */
	frame = gtk_frame_new(_("Show/Hide"));
	gtk_frame_set_label_align(GTK_FRAME(frame), 0.03, 0.5);
	gtk_container_set_border_width(GTK_CONTAINER(frame), 5);
	gtk_box_pack_start(GTK_BOX(main_vbox), frame, FALSE, FALSE, 2);
	gtk_widget_show(frame);

	vbox = gtk_vbox_new(FALSE, 0);
	gtk_container_set_border_width(GTK_CONTAINER(vbox), 4);
	gtk_container_add(GTK_CONTAINER(frame), vbox);
	gtk_widget_show(vbox);

	check = gtk_check_button_new_with_mnemonic(_("C_lose button"));
	prefsui->show_close_button = check;
	gtk_box_pack_start(GTK_BOX(vbox), check, FALSE, FALSE, 2);
	show = TRUE;
	kz_profile_get_value(kz_global_profile,
			     "Tab", "show_close_button",
			     &show, sizeof(show),
			     KZ_PROFILE_VALUE_TYPE_BOOL);
	gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(check), show);
	g_signal_connect(G_OBJECT(check), "toggled",
			 G_CALLBACK(cb_changed), prefsui);
	gtk_widget_show(check);

	check = gtk_check_button_new_with_mnemonic(_("_Progress bar"));
	prefsui->show_progressbar = check;
	gtk_box_pack_start(GTK_BOX(vbox), check, FALSE, FALSE, 2);
	show = FALSE;
	kz_profile_get_value(kz_global_profile,
			     "Tab", "show_progress",
			     &show, sizeof(show),
			     KZ_PROFILE_VALUE_TYPE_BOOL);
	gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(check), show);
	g_signal_connect(G_OBJECT(check), "toggled",
			 G_CALLBACK(cb_changed), prefsui);
	gtk_widget_show(check);

	prefsui->changed = FALSE;

	return main_vbox;
}


static void
prefs_tab_response (GtkWidget *widget, gint response)
{
	KzPrefsTab *prefsui = g_object_get_data(G_OBJECT(widget), DATA_KEY);

	g_return_if_fail(prefsui);

	switch (response) {
	case GTK_RESPONSE_ACCEPT:
	case GTK_RESPONSE_APPLY:
	{
		GtkAdjustment *adj;
		gint width;
		gboolean fix_width, show_button, show_progress;

		if (!prefsui->changed)
			break;

		fix_width = gtk_toggle_button_get_active
				(GTK_TOGGLE_BUTTON(prefsui->fix_width_check));
		kz_profile_set_value(kz_global_profile,
				     "Tab", "fix_width",
				     &fix_width, sizeof(fix_width),
				     KZ_PROFILE_VALUE_TYPE_BOOL);

		adj = gtk_spin_button_get_adjustment
			(GTK_SPIN_BUTTON(prefsui->width_spin));
		width = gtk_adjustment_get_value(adj);
		kz_profile_set_value(kz_global_profile,
				     "Tab", "fixed_width",
				     &width, sizeof(width),
				     KZ_PROFILE_VALUE_TYPE_INT);

		show_button = gtk_toggle_button_get_active
				(GTK_TOGGLE_BUTTON(prefsui->show_close_button));
		kz_profile_set_value(kz_global_profile,
				     "Tab", "show_close_button",
				     &show_button, sizeof(show_button),
				     KZ_PROFILE_VALUE_TYPE_BOOL);

		show_progress = gtk_toggle_button_get_active
				(GTK_TOGGLE_BUTTON(prefsui->show_progressbar));
		kz_profile_set_value(kz_global_profile,
				     "Tab", "show_progress",
				     &show_progress, sizeof(show_progress),
				     KZ_PROFILE_VALUE_TYPE_BOOL);

		prefsui->changed = FALSE;

		break;
	}
	case GTK_RESPONSE_REJECT:
		break;
	default:
		break;
	}
}


KzPrefsWinPageEntry *
prefs_tab_get_entry (gint idx)
{
	if (idx == 0)
		return &prefs_entry;
	else
		return NULL;
}
